@@ -17,17 +17,34 @@ Example:
 ...     frozenlist.filter(lambda x: x<10)
 ... )
 """
+from __future__ import annotations
 
 import builtins
 import functools
-from typing import Any, Callable, Iterable, List, Tuple, TypeVar, cast, overload
+import itertools
+from typing import (
+    Any,
+    Callable,
+    Generic,
+    Iterable,
+    Iterator,
+    List,
+    Optional,
+    Tuple,
+    TypeVar,
+    cast,
+    get_origin,
+    overload,
+)
 
-from expression.core import Matcher, Nothing, Option, Some, pipe
+from expression.core import Case, Nothing, Option, Some, pipe
 
 from . import seq
 
 TSource = TypeVar("TSource")
+TSourceIn = TypeVar("TSourceIn", contravariant=True)
 TResult = TypeVar("TResult")
+TResultOut = TypeVar("TResultOut", covariant=True)
 TState = TypeVar("TState")
 T1 = TypeVar("T1")
 T2 = TypeVar("T2")
@@ -37,48 +54,51 @@ T5 = TypeVar("T5")
 T6 = TypeVar("T6")
 
 
-class FrozenList(Tuple[TSource]):
+class FrozenList(Generic[TSource]):
     """Immutable list type.
 
+    Is faster than `List` for prepending, but slower for
+    appending.
+
+    Count: 200K
+
+    | Operation | FrozenList | List   |
+    |-----------|------------|--------|
+    | Append    | 3.29 s     | 0.02 s |
+    | Prepend   | 0.05 s     | 7.02 s |
+
     Example:
         >>> xs = Cons(5, Cons(4, Cons(3, Cons(2, Cons(1, Nil)))))
         >>> ys = empty.cons(1).cons(2).cons(3).cons(4).cons(5)
     """
 
-    @staticmethod
-    def of_seq(source: Iterable[TSource]) -> "FrozenList[TSource]":
-        return FrozenList(source)
-
-    @overload
-    def match(self) -> Matcher:
-        ...
-
-    @overload
-    def match(self, pattern: Any) -> Iterable[List[TSource]]:
-        ...
+    def __init__(self, value: Optional[Iterable[TSource]] = None) -> None:
+        # Use composition instead of inheritance since generic tuples
+        # are not suppored by mypy.
+        self.value: Tuple[TSource, ...] = tuple(value) if value else tuple()
 
     def match(self, pattern: Any) -> Any:
-        m = Matcher(self)
-        return m.case(pattern) if pattern else m
+        case: Case[TSource] = Case(self)
+        return case(pattern) if pattern else case
 
     @overload
-    def pipe(self, __fn1: Callable[["FrozenList[TSource]"], TResult]) -> TResult:
+    def pipe(self, __fn1: Callable[[FrozenList[TSource]], TResult]) -> TResult:
         ...
 
     @overload
-    def pipe(self, __fn1: Callable[["FrozenList[TSource]"], T1], __fn2: Callable[[T1], T2]) -> T2:
+    def pipe(self, __fn1: Callable[[FrozenList[TSource]], T1], __fn2: Callable[[T1], T2]) -> T2:
         ...
 
     @overload
     def pipe(
-        self, __fn1: Callable[["FrozenList[TSource]"], T1], __fn2: Callable[[T1], T2], __fn3: Callable[[T2], T3]
+        self, __fn1: Callable[[FrozenList[TSource]], T1], __fn2: Callable[[T1], T2], __fn3: Callable[[T2], T3]
     ) -> T3:
         ...
 
     @overload
     def pipe(
         self,
-        __fn1: Callable[["FrozenList[TSource]"], T1],
+        __fn1: Callable[[FrozenList[TSource]], T1],
         __fn2: Callable[[T1], T2],
         __fn3: Callable[[T2], T3],
         __fn4: Callable[[T3], T4],
@@ -88,7 +108,7 @@ class FrozenList(Tuple[TSource]):
     @overload
     def pipe(
         self,
-        __fn1: Callable[["FrozenList[TSource]"], T1],
+        __fn1: Callable[[FrozenList[TSource]], T1],
         __fn2: Callable[[T1], T2],
         __fn3: Callable[[T2], T3],
         __fn4: Callable[[T3], T4],
@@ -99,7 +119,7 @@ class FrozenList(Tuple[TSource]):
     @overload
     def pipe(
         self,
-        __fn1: Callable[["FrozenList[TSource]"], T1],
+        __fn1: Callable[[FrozenList[TSource]], T1],
         __fn2: Callable[[T1], T2],
         __fn3: Callable[[T2], T3],
         __fn4: Callable[[T3], T4],
@@ -112,10 +132,12 @@ class FrozenList(Tuple[TSource]):
         """Pipe list through the given functions."""
         return pipe(self, *args)
 
-    def append(self, other: "FrozenList[TSource]") -> "FrozenList[TSource]":
-        return FrozenList(self + other)
+    def append(self, other: FrozenList[TSource]) -> FrozenList[TSource]:
+        """Append frozen list to end of the frozen list."""
 
-    def choose(self, chooser: Callable[[TSource], Option[TResult]]) -> "FrozenList[TResult]":
+        return FrozenList(self.value + other.value)
+
+    def choose(self, chooser: Callable[[TSource], Option[TResult]]) -> FrozenList[TResult]:
         """Choose items from the list.
 
         Applies the given function to each element of the list. Returns
@@ -135,17 +157,23 @@ class FrozenList(Tuple[TSource]):
 
         return self.collect(mapper)
 
-    def collect(self, mapping: Callable[[TSource], "FrozenList[TResult]"]) -> "FrozenList[TResult]":
-        mapped = builtins.map(mapping, self)
+    def collect(self, mapping: Callable[[TSource], FrozenList[TResult]]) -> FrozenList[TResult]:
+        mapped = builtins.map(mapping, self.value)
         xs = (y for x in mapped for y in x)
         return FrozenList(xs)
 
-    def cons(self, element: TSource) -> "FrozenList[TSource]":
-        """Add element to front of List."""
+    def cons(self, element: TSource) -> FrozenList[TSource]:
+        """Add element to front of list."""
 
-        return FrozenList((element, *self))
+        return FrozenList((element,) + self.value)  # NOTE: Faster than (element, *self)
 
-    def filter(self, predicate: Callable[[TSource], bool]) -> "FrozenList[TSource]":
+    @staticmethod
+    def empty() -> FrozenList[TSource]:
+        """Returns empty list."""
+
+        return FrozenList()
+
+    def filter(self, predicate: Callable[[TSource], bool]) -> FrozenList[TSource]:
         """Filter list.
 
         Returns a new collection containing only the elements of the
@@ -158,7 +186,7 @@ class FrozenList(Tuple[TSource]):
             A list containing only the elements that satisfy the
             predicate.
         """
-        return FrozenList(builtins.filter(predicate, self))
+        return FrozenList(builtins.filter(predicate, self.value))
 
     def fold(self, folder: Callable[[TState, TSource], TState], state: TState) -> TState:
         """Applies a function to each element of the collection,
@@ -181,6 +209,18 @@ class FrozenList(Tuple[TSource]):
         """
         return functools.reduce(folder, self, state)
 
+    def forall(self, predicate: Callable[[TSource], bool]) -> bool:
+        """Tests if all elements of the collection satisfy the given
+        predicate.
+
+        Args:
+            predicate: The function to test the input elements.
+
+        Returns:
+            True if all of the elements satisfy the predicate.
+        """
+        return all(predicate(x) for x in self)
+
     def head(self) -> TSource:
         """Returns the first element of the list.
 
@@ -197,7 +237,7 @@ class FrozenList(Tuple[TSource]):
         head, *_ = self
         return head
 
-    def indexed(self, start: int = 0) -> "FrozenList[Tuple[int, TSource]]":
+    def indexed(self, start: int = 0) -> FrozenList[Tuple[int, TSource]]:
         """Returns a new list whose elements are the corresponding
         elements of the input list paired with the index (from `start`)
         of each element.
@@ -208,7 +248,7 @@ class FrozenList(Tuple[TSource]):
         Returns:
             The list of indexed elements.
         """
-        return FrozenList.of_seq(enumerate(self))
+        return of_seq(enumerate(self))
 
     def item(self, index: int) -> TSource:
         """Indexes into the list. The first element has index 0.
@@ -226,7 +266,7 @@ class FrozenList(Tuple[TSource]):
 
         return not bool(self)
 
-    def map(self, mapping: Callable[[TSource], TResult]) -> "FrozenList[TResult]":
+    def map(self, mapping: Callable[[TSource], TResult]) -> FrozenList[TResult]:
         """Map list.
 
         Builds a new collection whose elements are the results of
@@ -242,7 +282,61 @@ class FrozenList(Tuple[TSource]):
         """
         return FrozenList((*builtins.map(mapping, self),))
 
-    def skip(self, count: int) -> "FrozenList[TSource]":
+    def mapi(self, mapping: Callable[[int, TSource], TResult]) -> FrozenList[TResult]:
+        """Map list with index.
+
+        Builds a new collection whose elements are the results of
+        applying the given function to each of the elements of the
+        collection. The integer index passed to the function indicates
+        the index (from 0) of element being transformed.
+
+        Args:
+            mapping: The function to transform elements and their
+                indices.
+
+        Returns:
+            The list of transformed elements.
+        """
+        return FrozenList((*itertools.starmap(mapping, enumerate(self)),))
+
+    @staticmethod
+    def of(*args: TSource) -> FrozenList[TSource]:
+        """Create list from a number of arguments."""
+        return FrozenList((*args,))
+
+    @staticmethod
+    def of_seq(xs: Iterable[TSource]) -> FrozenList[TSource]:
+        """Create list from iterable sequence."""
+        return FrozenList((*xs,))
+
+    @staticmethod
+    def of_option(option: Option[TSource]) -> FrozenList[TSource]:
+        return of_option(option)
+
+    @overload
+    @staticmethod
+    def range(stop: int) -> FrozenList[int]:
+        ...
+
+    @overload
+    @staticmethod
+    def range(start: int, stop: int) -> FrozenList[int]:
+        ...
+
+    @overload
+    @staticmethod
+    def range(start: int, stop: int, step: int) -> FrozenList[int]:
+        ...
+
+    @staticmethod
+    def range(*args: int, **kw: int) -> FrozenList[int]:
+        return range(*args, **kw)
+
+    @staticmethod
+    def singleton(item: TSource) -> FrozenList[TSource]:
+        return singleton(item)
+
+    def skip(self, count: int) -> FrozenList[TSource]:
         """Returns the list after removing the first N elements.
 
         Args:
@@ -251,18 +345,18 @@ class FrozenList(Tuple[TSource]):
         Returns:
             The list after removing the first N elements.
         """
-        return FrozenList(self[count:])
+        return FrozenList(self.value[count:])
 
-    def skip_last(self, count: int) -> "FrozenList[TSource]":
-        return FrozenList(self[:-count])
+    def skip_last(self, count: int) -> FrozenList[TSource]:
+        return FrozenList(self.value[:-count])
 
-    def tail(self) -> "FrozenList[TSource]":
+    def tail(self) -> FrozenList[TSource]:
         """Return tail of List."""
 
-        _, *tail = self
+        _, *tail = self.value
         return FrozenList(tail)
 
-    def take(self, count: int) -> "FrozenList[TSource]":
+    def take(self, count: int) -> FrozenList[TSource]:
         """Returns the first N elements of the list.
 
         Args:
@@ -271,9 +365,9 @@ class FrozenList(Tuple[TSource]):
         Returns:
             The result list.
         """
-        return FrozenList(self[:count])
+        return FrozenList(self.value[:count])
 
-    def take_last(self, count: int) -> "FrozenList[TSource]":
+    def take_last(self, count: int) -> FrozenList[TSource]:
         """Returns a specified number of contiguous elements from the
         end of the list.
 
@@ -283,16 +377,20 @@ class FrozenList(Tuple[TSource]):
         Returns:
             The result list.
         """
-        return FrozenList(self[-count:])
+        return FrozenList(self.value[-count:])
 
     def try_head(self) -> Option[TSource]:
         """Returns the first element of the list, or None if the list is
         empty.
         """
-        return Some(self[0]) if self else Nothing
+        if self.value:
+            value = cast("TSource", self.value[0])
+            return Some(value)
+
+        return Nothing
 
     @staticmethod
-    def unfold(generator: Callable[[TState], Option[Tuple[TSource, TState]]], state: TState) -> "FrozenList[TSource]":
+    def unfold(generator: Callable[[TState], Option[Tuple[TSource, TState]]], state: TState) -> FrozenList[TSource]:
         """Returns a list that contains the elements generated by the
         given computation. The given initial state argument is passed to
         the element generator.
@@ -307,13 +405,9 @@ class FrozenList(Tuple[TSource]):
             The result list.
         """
 
-        result = generator(state)
-        for (item, state_) in result.to_list():
-            return FrozenList.unfold(generator, state_).cons(item)
-        else:
-            return empty
+        return pipe(state, unfold(generator))
 
-    def zip(self, other: "FrozenList[TResult]") -> "FrozenList[Tuple[TSource, TResult]]":
+    def zip(self, other: FrozenList[TResult]) -> FrozenList[Tuple[TSource, TResult]]:
         """Combines the two lists into a list of pairs. The two lists
         must have equal lengths. .
 
@@ -324,14 +418,38 @@ class FrozenList(Tuple[TSource]):
             A single list containing pairs of matching elements from the
             input lists.
         """
-        return FrozenList.of_seq(builtins.zip(self, other))
+        return of_seq(builtins.zip(self.value, other.value))
 
-    def __match__(self, pattern: Any) -> Iterable[TSource]:
+    def __add__(self, other: FrozenList[TSource]) -> FrozenList[TSource]:
+        return FrozenList(self.value + other.value)
+
+    @overload
+    def __getitem__(self, key: slice) -> FrozenList[TSource]:
+        ...
+
+    @overload
+    def __getitem__(self, key: int) -> TSource:
+        ...
+
+    def __getitem__(self, key: Any) -> Any:
+        return self.value[key]
+
+    def __iter__(self) -> Iterator[TSource]:
+        return iter(self.value)
+
+    def __eq__(self, o: Any) -> bool:
+        return self.value == o
+
+    def __len__(self) -> int:
+        return len(self.value)
+
+    def __match__(self, pattern: Any) -> Iterable[List[TSource]]:
         if self == pattern:
             return [[val for val in self]]
 
         try:
-            if isinstance(self, pattern):
+            origin: Any = get_origin(pattern)
+            if isinstance(self, origin or pattern):
                 return [[val for val in self]]
         except TypeError:
             pass
@@ -388,6 +506,8 @@ def collect(mapping: Callable[[TSource], FrozenList[TResult]]) -> Callable[[Froz
 
 
 def concat(sources: Iterable[FrozenList[TSource]]) -> FrozenList[TSource]:
+    """Concatenate sequence of FrozenList's"""
+
     def reducer(t: FrozenList[TSource], s: FrozenList[TSource]) -> FrozenList[TSource]:
         return t.append(s)
 
@@ -395,7 +515,7 @@ def concat(sources: Iterable[FrozenList[TSource]]) -> FrozenList[TSource]:
 
 
 def cons(head: TSource, tail: FrozenList[TSource]) -> FrozenList[TSource]:
-    return FrozenList(head, *tail)
+    return FrozenList((head, *tail))
 
 
 nil: FrozenList[Any] = FrozenList()
@@ -419,7 +539,7 @@ def filter(predicate: Callable[[TSource], bool]) -> Callable[[FrozenList[TSource
         collection for which the given predicate returns `True`
 
         Args:
-            source: The input list..
+            source: The input list.
 
         Returns:
             A list containing only the elements that satisfy the
@@ -456,6 +576,23 @@ def fold(folder: Callable[[TState, TSource], TState], state: TState) -> Callable
     return _fold
 
 
+def forall(predicate: Callable[[TSource], bool]) -> Callable[[FrozenList[TSource]], bool]:
+    """Tests if all elements of the collection satisfy the given
+    predicate.
+
+    Args:
+        predicate: The function to test the input elements.
+
+    Returns:
+        True if all of the elements satisfy the predicate.
+    """
+
+    def _forall(source: FrozenList[TSource]) -> bool:
+        return source.forall(predicate)
+
+    return _forall
+
+
 def head(source: FrozenList[TSource]) -> TSource:
     """Returns the first element of the list.
 
@@ -498,7 +635,7 @@ def item(index: int) -> Callable[[FrozenList[TSource]], TSource]:
     return _item
 
 
-def is_empty(source: FrozenList[TSource]) -> bool:
+def is_empty(source: FrozenList[Any]) -> bool:
     """Returns `True` if the list is empty, `False` otherwise."""
     return source.is_empty()
 
@@ -522,6 +659,33 @@ def map(mapper: Callable[[TSource], TResult]) -> Callable[[FrozenList[TSource]],
     return _map
 
 
+def mapi(mapper: Callable[[int, TSource], TResult]) -> Callable[[FrozenList[TSource]], FrozenList[TResult]]:
+    """Map list with index.
+
+    Builds a new collection whose elements are the results of
+    applying the given function to each of the elements of the
+    collection. The integer index passed to the function indicates
+    the index (from 0) of element being transformed.
+
+    Args:
+        mapping: The function to transform elements and their
+            indices.
+
+    Returns:
+        The list of transformed elements.
+    """
+
+    def _mapi(source: FrozenList[TSource]) -> FrozenList[TResult]:
+        return source.mapi(mapper)
+
+    return _mapi
+
+
+def of(*args: TSource) -> FrozenList[TSource]:
+    """Create list from a number of arguments."""
+    return FrozenList((*args,))
+
+
 def of_seq(xs: Iterable[TSource]) -> FrozenList[TSource]:
     """Create list from iterable sequence."""
     return FrozenList((*xs,))
@@ -533,6 +697,25 @@ def of_option(option: Option[TSource]) -> FrozenList[TSource]:
     return empty
 
 
+@overload
+def range(stop: int) -> FrozenList[int]:
+    ...
+
+
+@overload
+def range(start: int, stop: int) -> FrozenList[int]:
+    ...
+
+
+@overload
+def range(start: int, stop: int, step: int) -> FrozenList[int]:
+    ...
+
+
+def range(*args: int, **kw: int) -> FrozenList[int]:
+    return FrozenList((*builtins.range(*args, **kw),))
+
+
 def singleton(value: TSource) -> FrozenList[TSource]:
     return FrozenList((value,))
 
@@ -620,7 +803,7 @@ def try_head(source: FrozenList[TSource]) -> Option[TSource]:
     return source.try_head()
 
 
-def unfold(generator: Callable[[TState], Option[Tuple[TSource, TState]]], state: TState) -> "FrozenList[TSource]":
+def unfold(generator: Callable[[TState], Option[Tuple[TSource, TState]]]) -> Callable[[TState], FrozenList[TSource]]:
     """Returns a list that contains the elements generated by the
     given computation. The given initial state argument is passed to
     the element generator.
@@ -635,7 +818,11 @@ def unfold(generator: Callable[[TState], Option[Tuple[TSource, TState]]], state:
         The result list.
     """
 
-    return FrozenList.unfold(generator, state)
+    def _unfold(state: TState) -> FrozenList[TSource]:
+        xs = pipe(state, seq.unfold(generator))
+        return FrozenList(xs)
+
+    return _unfold
 
 
 def zip(other: FrozenList[TResult]) -> Callable[[FrozenList[TSource]], FrozenList[Tuple[TSource, TResult]]]:
@@ -681,6 +868,7 @@ __all__ = [
     "item",
     "is_empty",
     "map",
+    "mapi",
     "of_seq",
     "of_option",
     "singleton",
