@@ -14,16 +14,17 @@
 # - Copyright (c) Microsoft Corporation. All Rights Reserved.
 # - MIT License
 # - https://github.com/fsharp/fsharp/blob/master/src/fsharp/FSharp.Core/map.fs
+from __future__ import annotations
 
-from typing import Any, Callable, Generic, Iterable, Iterator, List, Optional, Tuple, TypeVar, Union, cast, overload
+from typing import Any, Callable, Iterable, Iterator, List, Mapping, Optional, Set, Tuple, TypeVar, cast, overload
 
-from expression.core import Option, pipe
+from expression.core import Option, SupportsLessThan, pipe
 
 from . import maptree, seq
 from .frozenlist import FrozenList
 from .maptree import MapTree
 
-Key = TypeVar("Key")
+Key = TypeVar("Key", bound=SupportsLessThan)
 Value = TypeVar("Value")
 Result = TypeVar("Result")
 
@@ -35,31 +36,31 @@ T5 = TypeVar("T5")
 T6 = TypeVar("T6")
 
 
-class Map(Generic[Key, Value]):
+class Map(Mapping[Key, Value]):
     """The immutable map class."""
 
-    def __init__(self, __tree: Optional[MapTree[Key, Value]] = None, **kw: Value) -> None:
-        tree: MapTree[str, Value] = maptree.of_seq(kw.items())
-        self._tree = __tree if __tree is not None else tree
+    def __init__(self, __tree: Optional[MapTree[Key, Value]] = None) -> None:
+        self._tree: MapTree[Key, Value] = __tree if __tree else maptree.empty
+
+    def add(self, key: Key, value: Value) -> Map[Key, Value]:
+        return Map(maptree.add(key, value, self._tree))
 
     @overload
-    def pipe(self, __fn1: Callable[["Map[Key, Value]"], Result]) -> Result:
+    def pipe(self, __fn1: Callable[[Map[Key, Value]], Result]) -> Result:
         ...
 
     @overload
-    def pipe(self, __fn1: Callable[["Map[Key, Value]"], T1], __fn2: Callable[[T1], T2]) -> T2:
+    def pipe(self, __fn1: Callable[[Map[Key, Value]], T1], __fn2: Callable[[T1], T2]) -> T2:
         ...
 
     @overload
-    def pipe(
-        self, __fn1: Callable[["Map[Key, Value]"], T1], __fn2: Callable[[T1], T2], __fn3: Callable[[T2], T3]
-    ) -> T3:
+    def pipe(self, __fn1: Callable[[Map[Key, Value]], T1], __fn2: Callable[[T1], T2], __fn3: Callable[[T2], T3]) -> T3:
         ...
 
     @overload
     def pipe(
         self,
-        __fn1: Callable[["Map[Key, Value]"], T1],
+        __fn1: Callable[[Map[Key, Value]], T1],
         __fn2: Callable[[T1], T2],
         __fn3: Callable[[T2], T3],
         __fn4: Callable[[T3], T4],
@@ -69,7 +70,7 @@ class Map(Generic[Key, Value]):
     @overload
     def pipe(
         self,
-        __fn1: Callable[["Map[Key, Value]"], T1],
+        __fn1: Callable[[Map[Key, Value]], T1],
         __fn2: Callable[[T1], T2],
         __fn3: Callable[[T2], T3],
         __fn4: Callable[[T3], T4],
@@ -80,7 +81,7 @@ class Map(Generic[Key, Value]):
     @overload
     def pipe(
         self,
-        __fn1: Callable[["Map[Key, Value]"], T1],
+        __fn1: Callable[[Map[Key, Value]], T1],
         __fn2: Callable[[T1], T2],
         __fn3: Callable[[T2], T3],
         __fn4: Callable[[T3], T4],
@@ -94,29 +95,26 @@ class Map(Generic[Key, Value]):
         return pipe(self, *args)
 
     @staticmethod
-    def empty() -> "Map[Key, Value]":
-        return Map(maptree.empty)
+    def create(ie: Iterable[Tuple[Key, Value]]) -> Map[Key, Value]:
+        return create(ie)
+
+    def contains_key(self, key: Key) -> bool:
+        return maptree.mem(key, self._tree)
+
+    def change(self, key: Key, f: Callable[[Option[Value]], Option[Value]]) -> Map[Key, Value]:
+        return Map(maptree.change(key, f, self._tree))
 
     @staticmethod
-    def create(ie: Iterable[Tuple[Key, Value]]) -> "Map[Key, Value]":
-        return Map(maptree.of_seq(ie))
-
-    def add(self, key: Key, value: Value) -> "Map[Key, Value]":
-        return Map(maptree.add(key, value, self._tree))
-
-    def change(self, key: Key, f: Callable[[Option[Value]], Option[Value]]) -> "Map[Key, Value]":
-        return Map(maptree.change(key, f, self._tree))
+    def empty() -> Map[Key, Value]:
+        return Map(maptree.empty)
 
     def is_empty(self) -> bool:
         return maptree.is_empty(self._tree)
 
-    def try_pick(self, chooser: Callable[[Key, Value], Option[Result]]) -> Option[Result]:
-        return maptree.try_pick(chooser, self._tree)
-
     def exists(self, predicate: Callable[[Key, Value], bool]) -> bool:
         return maptree.exists(predicate, self._tree)
 
-    def filter(self, predicate: Callable[[Key, Value], bool]) -> "Map[Key, Value]":
+    def filter(self, predicate: Callable[[Key, Value], bool]) -> Map[Key, Value]:
         return Map(maptree.filter(predicate, self._tree))
 
     def for_all(self, predicate: Callable[[Key, Value], bool]) -> bool:
@@ -141,7 +139,10 @@ class Map(Generic[Key, Value]):
     def fold(self, folder: Callable[[Result, Tuple[Key, Value]], Result], state: Result) -> Result:
         return maptree.fold(folder, state, self._tree)
 
-    def map(self, mapping: Callable[[Value], Result]) -> "Map[Key, Result]":
+    def fold_back(self, folder: Callable[[Tuple[Key, Value], Result], Result], state: Result) -> Result:
+        return maptree.fold_back(folder, self._tree, state)
+
+    def map(self, mapping: Callable[[Key, Value], Result]) -> Map[Key, Result]:
         """Builds a new collection whose elements are the results of
         applying the given function to each of the elements of the
         collection. The key passed to the function indicates the key of
@@ -155,16 +156,41 @@ class Map(Generic[Key, Value]):
         """
         return Map(maptree.map(mapping, self._tree))
 
-    def partition(self, predicate: Callable[[Key, Value], bool]) -> "Tuple[Map[Key, Value], Map[Key, Value]]":
+    def partition(self, predicate: Callable[[Key, Value], bool]) -> Tuple[Map[Key, Value], Map[Key, Value]]:
         r1, r2 = maptree.partition(predicate, self._tree)
         return Map(r1), Map(r2)
 
-    def contains_key(self, key: Key) -> bool:
-        return maptree.mem(key, self._tree)
+    # @overload
+    # def get(self, key: Key) -> Optional[Value]:
+    #    ...
 
-    def remove(self, key: Key) -> "Map[Key, Value]":
+    # @overload
+    # def get(self, key: Key, default: Value) -> Value:
+    #    ...
+
+    # def get(self, key: Key, default: Union[Value, _T]) -> Union[Value, _T]:
+    #    for value in self.try_find(key):
+    #        return value
+
+    #   return default
+
+    def items(self) -> Set[Tuple[Key, Value]]:
+        return set(maptree.to_seq(self._tree))
+
+    def remove(self, key: Key) -> Map[Key, Value]:
         return Map(maptree.remove(key, self._tree))
 
+    def to_list(self) -> FrozenList[Tuple[Key, Value]]:
+        return maptree.to_list(self._tree)
+
+    def to_seq(self) -> Iterable[Tuple[Key, Value]]:
+        """Convert to sequence.
+
+        Returns:
+            Sequence of key, value tuples.
+        """
+        return maptree.to_seq(self._tree)
+
     def try_get_value(self, key: Key, value: List[Value]):
         for v in maptree.try_find(key, self._tree).to_list():
             value.append(v)
@@ -175,61 +201,72 @@ class Map(Generic[Key, Value]):
     def try_find(self, key: Key) -> Option[Value]:
         return maptree.try_find(key, self._tree)
 
-    def to_list(self) -> FrozenList[Tuple[Key, Value]]:
-        return maptree.to_list(self._tree)
-
-    def to_seq(self) -> Iterable[Tuple[Key, Value]]:
-        """Convert to sequence.
-
-        Returns:
-            Sequenc of key, value tuples.
-        """
-        return maptree.to_seq(self._tree)
-
-    @overload
-    @staticmethod
-    def of_list(lst: FrozenList[Tuple[Key, Value]]) -> "Map[Key, Value]":
-        ...
-
-    @overload
-    @staticmethod
-    def of_list(lst: List[Tuple[Key, Value]]) -> "Map[Key, Value]":
-        ...
+    def try_pick(self, chooser: Callable[[Key, Value], Option[Result]]) -> Option[Result]:
+        return maptree.try_pick(chooser, self._tree)
 
     @staticmethod
-    def of_list(lst: Union[List[Tuple[Key, Value]], FrozenList[Tuple[Key, Value]]]) -> "Map[Key, Value]":
+    def of(**args: Value) -> Map[str, Value]:
+        return Map(maptree.of_seq(args.items()))
+
+    @staticmethod
+    def of_frozenlist(lst: FrozenList[Tuple[Key, Value]]) -> Map[Key, Value]:
         """Generate map from list.
 
         Returns:
-            New map.
+            The new map.
         """
-        return Map(maptree.of_list(FrozenList(lst)))
+        return of_frozenlist((lst))
+
+    @staticmethod
+    def of_list(lst: List[Tuple[Key, Value]]) -> Map[Key, Value]:
+        """Generate map from list.
+
+        Returns:
+            The new map.
+        """
+        return of_list((lst))
+
+    @staticmethod
+    def of_seq(sequence: Iterable[Tuple[Key, Value]]) -> Map[Key, Value]:
+        """Generate map from sequence.
+
+        Generates a new map from an iterable of key/value tuples. This
+        is an alias for `Map.create`.
+
+        Returns:
+            The new map.
+        """
+        return of_seq(sequence)
 
     def __hash__(self) -> int:
         def combine_hash(x: int, y: int) -> int:
             return (x << 1) + y + 631
 
         res = 0
-        for x, y in self:
+        for x, y in maptree.mk_iterator(self._tree):
             res = combine_hash(res, hash(x))
             res = combine_hash(res, hash(y))
         return res
 
-    def __getitem__(self, key: Value) -> Value:
-        return maptree.find(key, self._tree)
+    def __getitem__(self, k: Key) -> Value:
+        return maptree.find(k, self._tree)
 
-    def __iter__(self) -> Iterator[Tuple[Key, Value]]:
-        return maptree.mk_iterator(self._tree)
+    def __iter__(self) -> Iterator[Key]:
+        xs = maptree.mk_iterator(self._tree)
+        return (k for (k, _) in xs)
 
     def __len__(self) -> int:
         """Return the number of bindings in the map."""
         return maptree.size(self._tree)
 
-    def __eq__(self, other: Any) -> bool:
-        if not isinstance(other, Map):
+    def __contains__(self, o: Any) -> bool:
+        return self.contains_key(o)
+
+    def __eq__(self, o: Any) -> bool:
+        if not isinstance(o, Map):
             return False
 
-        other = cast(Map[Any, Any], other)
+        other = cast(Map[Any, Any], o)
         iterator: Iterator[Tuple[Any, Any]] = iter(other.to_seq())
 
         for kv in self.to_seq():
@@ -248,10 +285,11 @@ class Map(Generic[Key, Value]):
     def __str__(self) -> str:
         def to_str(item: Tuple[Key, Value]) -> str:
             key, value = item
-            key = f'"{key}"' if isinstance(key, str) else key
+            if isinstance(key, str):
+                return f'("{key}", {value})'
             return f"({key}, {value})"
 
-        items = pipe(self, Map.to_seq, seq.map(to_str))
+        items = pipe(self.to_seq(), seq.map(to_str))
         return f"map [{'; '.join(items)}]"
 
     def __repr__(self) -> str:
@@ -313,11 +351,22 @@ def change(key: Key, fn: Callable[[Option[Value]], Option[Value]]) -> Callable[[
     return _change
 
 
-def count(table: Map[Key, Value]) -> int:
+def contains_key(key: Key) -> Callable[[Map[Key, Any]], bool]:
+    def _contains_key(table: Map[Key, Any]) -> bool:
+        return table.contains_key(key)
+
+    return _contains_key
+
+
+def count(table: Map[Any, Any]) -> int:
     """Return the number of bindings in the map."""
     return len(table)
 
 
+def create(ie: Iterable[Tuple[Key, Value]]) -> Map[Key, Value]:
+    return Map(maptree.of_seq(ie))
+
+
 def find(key: Key) -> Callable[[Map[Key, Value]], Value]:
     """Lookup an element in the map, raising KeyNotFoundException if no
     binding exists in the map
@@ -334,7 +383,7 @@ def find(key: Key) -> Callable[[Map[Key, Value]], Value]:
     return _find
 
 
-def is_empty(table: Map[Key, Value]) -> bool:
+def is_empty(table: Map[Any, Any]) -> bool:
     """Is the map empty?
 
     Args:
@@ -346,13 +395,6 @@ def is_empty(table: Map[Key, Value]) -> bool:
     return table.is_empty()
 
 
-def contains_key(key: Key) -> Callable[[Map[Key, Value]], bool]:
-    def _contains_key(table: Map[Key, Value]) -> bool:
-        return table.contains_key(key)
-
-    return _contains_key
-
-
 def iterate(action: Callable[[Key, Value], None]) -> Callable[[Map[Key, Value]], None]:
     def _iterate(table: Map[Key, Value]) -> None:
         return table.iterate(action)
@@ -378,8 +420,8 @@ def try_pick(chooser: Callable[[Key, Value], Option[Result]]) -> Callable[[Map[K
     return _try_pick
 
 
-def pick(chooser: Callable[[Key, Value], Option[Result]]) -> Callable[[Map[Key, Value]], Option[Result]]:
-    def _try_pick(table: Map[Key, Value]) -> Option[Result]:
+def pick(chooser: Callable[[Key, Value], Option[Result]]) -> Callable[[Map[Key, Value]], Result]:
+    def _try_pick(table: Map[Key, Value]) -> Result:
         for res in table.try_pick(chooser):
             return res
         else:
@@ -420,7 +462,7 @@ def for_all(predicate: Callable[[Key, Value], bool]) -> Callable[[Map[Key, Value
     return _for_all
 
 
-def map(mapping: Callable[[Value], Result]) -> Callable[[Map[Key, Value]], Map[Key, Result]]:
+def map(mapping: Callable[[Key, Value], Result]) -> Callable[[Map[Key, Value]], Map[Key, Result]]:
     def _map(table: Map[Key, Value]) -> Map[Key, Result]:
         return table.map(mapping)
 
@@ -434,9 +476,13 @@ def fold(folder: Callable[[Result, Tuple[Key, Value]], Result], state: Result) -
     return _fold
 
 
-# // [<CompiledName("FoldBack")>]
-# let foldBack<'Key, 'T, 'State  when 'Key : comparison> folder (table: Map<'Key, 'T>) (state:'State) =
-#     maptree.foldBack folder table.Tree state
+def fold_back(
+    folder: Callable[[Tuple[Key, Value], Result], Result], table: Map[Key, Value]
+) -> Callable[[Result], Result]:
+    def _fold_back(state: Result) -> Result:
+        return table.fold_back(folder, state)
+
+    return _fold_back
 
 
 def partition(
@@ -466,10 +512,6 @@ def remove(key: Key) -> Callable[[Map[Key, Value]], Map[Key, Value]]:
     return _remove
 
 
-def to_seq(table: Map[Key, Value]):
-    return Map.to_seq(table)
-
-
 # // [<CompiledName("FindKey")>]
 # let findKey predicate (table : Map<_, _>) =
 #     table |> Seq.pick (fun kvp -> let k = kvp.Key in if predicate k kvp.Value then Some k else None)
@@ -479,28 +521,31 @@ def to_seq(table: Map[Key, Value]):
 #     table |> Seq.tryPick (fun kvp -> let k = kvp.Key in if predicate k kvp.Value then Some k else None)
 
 
-@overload
-def of_list(elements: FrozenList[Tuple[Key, Value]]) -> Map[Key, Value]:
-    ...
+def of(**args: Value) -> Map[str, Value]:
+    """Create map from arguments."""
+    return Map(maptree.of_seq(args.items()))
+
+
+def of_frozenlist(elements: FrozenList[Tuple[Key, Value]]) -> Map[Key, Value]:
+    return Map(maptree.of_list(elements))
 
 
-@overload
 def of_list(elements: List[Tuple[Key, Value]]) -> Map[Key, Value]:
-    ...
-
-
-def of_list(elements: Union[List[Tuple[Key, Value]], FrozenList[Tuple[Key, Value]]]) -> Map[Key, Value]:
-    return Map.of_list(FrozenList(elements))
+    return Map(maptree.of_list(FrozenList(elements)))
 
 
 def of_seq(elements: Iterable[Tuple[Key, Value]]) -> Map[Key, Value]:
-    return Map.create(elements)
+    return Map(maptree.of_seq(elements))
 
 
 def to_list(table: Map[Key, Value]) -> FrozenList[Tuple[Key, Value]]:
     return table.to_list()
 
 
+def to_seq(table: Map[Key, Value]) -> Iterable[Tuple[Key, Value]]:
+    return table.to_seq()
+
+
 def try_find(key: Key) -> Callable[[Map[Key, Value]], Option[Value]]:
     """Lookup an element in the map, returning a `Some` value if the
     element is in the domain of the map and `Nothing` if not.
@@ -513,7 +558,7 @@ def try_find(key: Key) -> Callable[[Map[Key, Value]], Option[Value]]:
         instance and returns the result.
     """
 
-    def _try_find(table: Map[Key, Value]):
+    def _try_find(table: Map[Key, Value]) -> Option[Value]:
         """Lookup an element in the map, returning a `Some` value if the
         element is in the domain of the map and `Nothing` if not.
 
@@ -530,35 +575,11 @@ def try_find(key: Key) -> Callable[[Map[Key, Value]], Option[Value]]:
 
 empty: Map[Any, Any] = Map.empty()
 
-# let groupBy (projection: 'T -> 'Key) (xs: 'T seq) ([<Fable.Core.Inject>] comparer: IEqualityComparer<'Key>): ('Key * 'T seq) seq =
-#     let dict: Fable.Core.JS.Map<_,ResizeArray<'T>> = createMutable Seq.empty comparer
-
-#     // Build the groupings
-#     for v in xs do
-#         let key = projection v
-#         if dict.has(key) then dict.get(key).Add(v)
-#         else dict.set(key, ResizeArray [v]) |> ignore
-
-#     // Mapping shouldn't be necessary because KeyValuePair compiles
-#     // as a tuple, but let's do it just in case the implementation changes
-#     dict.entries() |> Seq.map (fun (k,v) -> k, upcast v)
-
-# let countBy (projection: 'T -> 'Key) (xs: 'T seq) ([<Fable.Core.Inject>] comparer: IEqualityComparer<'Key>): ('Key * int) seq =
-#     let dict = createMutable Seq.empty comparer
-
-#     for value in xs do
-#         let key = projection value
-#         if dict.has(key) then dict.set(key, dict.get(key) + 1)
-#         else dict.set(key, 1)
-#         |> ignore
-
-#     dict.entries()
-
-
 __all__ = [
     "Map",
     "add",
     "change",
+    "create",
     "contains_key",
     "count",
     "empty",
@@ -570,7 +591,10 @@ __all__ = [
     "is_empty",
     "iterate",
     "map",
+    "of",
+    "of_frozenlist",
     "of_list",
+    "of_seq",
     "partition",
     "pick",
     "remove",
