@@ -26,12 +26,12 @@ import builtins
 from dataclasses import dataclass
 from typing import Any, Callable, Generic, Iterable, Iterator, Tuple, TypeVar, cast
 
-from expression.core import Nothing, Option, Some, failwith, pipe
+from expression.core import Nothing, Option, Some, SupportsLessThan, failwith, pipe
 
 from . import frozenlist, seq
 from .frozenlist import FrozenList
 
-Key = TypeVar("Key")
+Key = TypeVar("Key", bound=SupportsLessThan)
 Value = TypeVar("Value")
 Result = TypeVar("Result")
 
@@ -56,11 +56,11 @@ class MapTreeNode(MapTreeLeaf[Key, Value]):
 empty: MapTree[Any, Any] = Nothing
 
 
-def is_empty(m: MapTree[Key, Value]):
+def is_empty(m: MapTree[Any, Any]):
     return m.is_none()
 
 
-def size_aux(acc: int, m: MapTree[Key, Value]) -> int:
+def size_aux(acc: int, m: MapTree[Any, Any]) -> int:
     for m2 in m.to_list():
         if isinstance(m2, MapTreeNode):
             mn = cast(MapTreeNode[Key, Value], m2)
@@ -71,11 +71,11 @@ def size_aux(acc: int, m: MapTree[Key, Value]) -> int:
         return acc
 
 
-def size(x: MapTree[Key, Value]):
+def size(x: MapTree[Any, Any]):
     return size_aux(0, x)
 
 
-def height(m: MapTree[Key, Value]) -> int:
+def height(m: MapTree[Any, Any]) -> int:
     for m2 in m.to_list():
         if isinstance(m2, MapTreeNode):
             mn = cast(MapTreeNode[Key, Value], m2)
@@ -150,7 +150,8 @@ def add(k: Key, v: Value, m: MapTree[Key, Value]) -> MapTree[Key, Value]:
                 return rebalance(mn.left, mn.key, mn.value, add(k, v, mn.right))
         else:
             if k < m2.key:
-                return Some(MapTreeNode(k, v, empty, m, 2))
+                node = MapTreeNode(k, v, empty, m, 2)
+                return Some(node)
             elif k == m2:
                 return Some(MapTreeLeaf(k, v))
             else:
@@ -166,7 +167,7 @@ def try_find(k: Key, m: MapTree[Key, Value]) -> Option[Value]:
         else:
             if isinstance(m2, MapTreeNode):
                 mn = cast(MapTreeNode[Key, Value], m2)
-                return try_find(k, mn.left if k < m2 else mn.right)
+                return try_find(k, mn.left if k < mn.key else mn.right)
             else:
                 return Nothing
     else:  # Nothing
@@ -183,9 +184,10 @@ def find(k: Key, m: MapTree[Key, Value]) -> Value:
 def partition1(
     predicate: Callable[[Key, Value], bool], k: Key, v: Value, acc: Tuple[MapTree[Key, Value], MapTree[Key, Value]]
 ) -> Tuple[MapTree[Key, Value], MapTree[Key, Value]]:
-    (acc1, acc2) = acc
+    acc1, acc2 = acc
     if predicate(k, v):
-        return add(k, v, acc1), acc2
+        a: MapTree[Key, Value] = add(k, v, acc1)
+        return a, acc2
     else:
         return acc1, add(k, v, acc2)
 
@@ -322,7 +324,7 @@ def change(k: Key, u: Callable[[Option[Value]], Option[Value]], m: MapTree[Key,
             return m
 
 
-def mem(k: Key, m: MapTree[Key, Value]) -> bool:
+def mem(k: Key, m: MapTree[Key, Any]) -> bool:
     for m2 in m.to_list():
         if isinstance(m2, MapTreeNode):
             mn = cast(MapTreeNode[Key, Value], m2)
@@ -352,10 +354,12 @@ def try_pick(f: Callable[[Key, Value], Option[Result]], m: MapTree[Key, Value])
     for m2 in m.to_list():
         if isinstance(m2, MapTreeNode):
             mn = cast(MapTreeNode[Key, Value], m2)
-            for res in try_pick(f, mn.left).to_list():
+            res = try_pick(f, mn.left)
+            if res.is_some():
                 return res
             else:
-                for res in f(mn.key, mn.value):
+                res = f(mn.key, mn.value)
+                if res.is_some():
                     return res
                 else:
                     return try_pick(f, mn.right)
@@ -387,30 +391,16 @@ def forall(f: Callable[[Key, Value], bool], m: MapTree[Key, Value]) -> bool:
         return True
 
 
-def map(f: Callable[[Value], Result], m: MapTree[Key, Value]) -> MapTree[Key, Result]:
+def map(f: Callable[[Key, Value], Result], m: MapTree[Key, Value]) -> MapTree[Key, Result]:
     for m2 in m.to_list():
         if isinstance(m2, MapTreeNode):
             mn = cast(MapTreeNode[Key, Value], m2)
             l2 = map(f, mn.left)
-            v2 = f(mn.value)
+            v2 = f(mn.key, mn.value)
             r2 = map(f, mn.right)
             return Some(MapTreeNode(mn.key, v2, l2, r2, mn.height))
         else:
-            return Some(MapTreeLeaf(m2.key, f(m2.value)))
-    else:
-        return empty
-
-
-def mapi(f: Callable[[Tuple[Key, Value]], Result], m: MapTree[Key, Value]) -> MapTree[Key, Value]:
-    for m2 in m.to_list():
-        if isinstance(m2, MapTreeNode):
-            mn = cast(MapTreeNode[Key, Value], m2)
-            l2 = mapi(f, mn.left)
-            v2 = f((mn.key, mn.value))
-            r2 = mapi(f, mn.right)
-            return Some(MapTreeNode(mn.key, v2, l2, r2, mn.height))
-        else:
-            return Some(MapTreeLeaf(m2.key, f((m2.key, m2.value))))
+            return Some(MapTreeLeaf(m2.key, f(m2.key, m2.value)))
     else:
         return empty
 
@@ -473,12 +463,8 @@ def mk_from_iterator(acc: MapTree[Key, Value], e: Iterator[Tuple[Key, Value]]) -
 
 
 def of_seq(xs: Iterable[Tuple[Key, Value]]) -> MapTree[Key, Value]:
-    if isinstance(xs, FrozenList):
-        xs = cast(FrozenList[Tuple[Key, Value]], xs)
-        return of_list(xs)
-    else:
-        ie = builtins.iter(xs)
-        return mk_from_iterator(empty, ie)
+    ie = builtins.iter(xs)
+    return mk_from_iterator(empty, ie)
 
 
 # Imperative left-to-right iterators.
