@@ -1,3 +1,8 @@
+"""The system module.
+
+Contains tools and utilities for dealing with (async) disposables and
+cancellation tokens.
+"""
 from .cancellation import CancellationToken, CancellationTokenSource
 from .disposable import (
     AnonymousDisposable,
