@@ -1,11 +1,12 @@
 from typing import Any, Callable, Generator
 
 import pytest
-from expression import effect
-from expression.core import Nothing, Option, Some, option, pipe, pipe2
 from hypothesis import given
 from hypothesis import strategies as st
 
+from expression import effect
+from expression.core import Nothing, Option, Some, match, option, pipe, pipe2
+from expression.extra.option import pipeline
 from tests.utils import CustomException
 
 
@@ -17,6 +18,17 @@ def test_option_some():
     assert pipe(xs, option.is_none) is False
 
 
+def test_option_some_match():
+    xs = Some(42)
+
+    with match(xs) as case:
+        for x in case(Some[int]):
+            assert x == 42
+
+        while case.default():
+            assert False
+
+
 def test_option_some_iterate():
     xs = Some(42)
 
@@ -35,6 +47,20 @@ def test_option_none():
     assert xs.pipe(option.is_none) is True
 
 
+def test_option_none_match():
+    xs = Nothing
+
+    with match(xs) as case:
+        for _ in case(Some[int]):
+            assert False
+
+        while case(Nothing):
+            assert True
+
+        while case.default():
+            assert False
+
+
 def test_option_nothing_iterate():
     xs = Nothing
 
@@ -57,7 +83,38 @@ def test_option_none_not_equals_some():
     assert ys != xs
 
 
-@given(st.one_of(st.integers(), st.text(), st.floats()), st.one_of(st.integers(), st.text(), st.floats()))
+def test_option_none_default_value():
+    xs = Nothing
+
+    zs = xs.default_value(42)
+
+    assert zs == 42
+
+
+def test_option_some_default_value():
+    xs: Option[int] = Some(42)
+    zs = xs.default_value(0)
+
+    assert zs == 42
+
+
+def test_option_none_default_arg():
+    xs = Nothing
+    zs = option.default_arg(xs, 42)
+
+    assert zs == 42
+
+
+def test_option_some_default_arg():
+    xs: Option[int] = Some(42)
+    zs = option.default_arg(xs, 0)
+
+    assert zs == 42
+
+
+@given(
+    st.one_of(st.integers(), st.text(), st.floats()), st.one_of(st.integers(), st.text(), st.floats())  # type: ignore
+)
 def test_option_some_equals_some(a: Any, b: Any):
     xs = Some(a)
     ys = Some(b)
@@ -164,6 +221,17 @@ def test_option_some_bind_piped():
         assert False
 
 
+def test_option_filter_none():
+    xs = Nothing
+    assert xs.filter(lambda x: True) == Nothing
+
+
+def test_option_filter_some():
+    xs = Some(42)
+    assert xs.filter(lambda x: x > 41) == Some(42)
+    assert xs.filter(lambda x: x > 42) == Nothing
+
+
 def test_option_none_to_list():
     xs = Nothing
     assert xs.to_list() == []
@@ -247,6 +315,19 @@ def test_option_builder_yield_value():
         assert False
 
 
+def test_option_builder_yield_value_async():
+    @effect.option
+    def fn():
+        yield 42
+
+    xs = fn()
+    for value in xs.match(Some):
+        assert value == 42
+        break
+    else:
+        assert False
+
+
 def test_option_builder_yield_some_wrapped():
     @effect.option
     def fn() -> Generator[Option[int], Option[int], Option[int]]:
@@ -306,6 +387,7 @@ def test_option_builder_yield_from_some():
 def test_option_builder_yield_from_none():
     @effect.option
     def fn() -> Generator[int, int, int]:
+        x: int
         x = yield from Nothing
         return x
 
@@ -332,6 +414,7 @@ def test_option_builder_multiple_some():
 def test_option_builder_none_short_circuits():
     @effect.option
     def fn() -> Generator[int, int, int]:
+        x: int
         x = yield from Nothing
         y = yield from Some(43)
 
@@ -353,3 +436,34 @@ def test_option_builder_throws():
         fn()
 
     assert ex.value.message == error
+
+
+def test_pipeline_none():
+
+    hn = pipeline()
+
+    assert hn(42) == Some(42)
+
+
+def test_pipeline_works():
+    fn: Callable[[int], Option[int]] = lambda x: Some(x * 10)
+    gn: Callable[[int], Option[int]] = lambda x: Some(x + 10)
+
+    hn = pipeline(
+        fn,
+        gn,
+    )
+
+    assert hn(42) == Some(430)
+
+
+def test_pipeline_error():
+    fn: Callable[[int], Option[int]] = lambda x: Some(x * 10)
+    gn: Callable[[int], Option[int]] = lambda x: Nothing
+
+    hn = pipeline(
+        fn,
+        gn,
+    )
+
+    assert hn(42) == Nothing
