@@ -1,12 +1,13 @@
-from typing import Callable, Generator, List, Optional
+from typing import Any, Callable, Generator, List, Optional
 
 import pytest
-from expression import effect
-from expression.core import Error, Ok, Result, result
-from expression.extra.result import sequence
 from hypothesis import given
 from hypothesis import strategies as st
 
+from expression import effect
+from expression.core import Error, Ok, Result, match, result
+from expression.extra.result import pipeline, sequence
+
 from .utils import CustomException
 
 
@@ -18,13 +19,35 @@ def test_result_ok():
     assert not xs.is_error()
     assert str(xs) == "Ok 42"
 
-    for x in xs.match(Ok):
+    for x in xs.match(Ok[int, str]):
         assert x == 42
         break
     else:
         assert False
 
 
+def test_result_match_ok():
+    xs: Result[int, str] = Ok(42)
+
+    with match(xs) as case:
+        for x in case(Ok[int, str]):
+            assert x == 42
+            break
+        else:
+            assert False
+
+
+def test_result_match_error():
+    xs: Result[int, str] = Error("err")
+
+    with match(xs) as case:
+        for err in case(Error[int, str]):
+            assert err == "err"
+            break
+        else:
+            assert False
+
+
 def test_result_ok_iterate():
     for x in Ok(42):
         assert x == 42
@@ -39,10 +62,10 @@ def test_result_error():
     assert xs.is_error()
     assert str(xs) == f"Error {error}"
 
-    for _ in xs.match(Ok):
+    for _ in xs.match(Ok[str, Exception]):
         assert False
 
-    for ex in xs.match(Error):
+    for ex in xs.match(Error[str, Exception]):
         assert ex == error
 
 
@@ -82,8 +105,8 @@ def test_result_map_piped(x: int, y: int):
     xs: Result[int, Exception] = Ok(x)
     mapper: Callable[[int], int] = lambda x: x + y
 
-    ys = xs.pipe(result.map(mapper))
-    for value in ys.match(Ok):
+    ys = xs.pipe(result.map(mapper))  # NOTE: shows type error for mypy
+    for value in ys.match(Ok[int, Exception]):
         assert value == mapper(x)
         break
     else:
@@ -96,7 +119,7 @@ def test_result_map_ok_fluent(x: int, y: int):
     mapper: Callable[[int], int] = lambda x: x + y
 
     ys = xs.map(mapper)
-    for value in ys.match(Ok):
+    for value in ys.match(Ok[int, Exception]):
         assert value == mapper(x)
         break
     else:
@@ -111,7 +134,7 @@ def test_result_ok_chained_map(x: int, y: int):
 
     ys = xs.map(mapper1).map(mapper2)
 
-    for value in ys.match(Ok):
+    for value in ys.match(Ok[int, Exception]):
         assert value == mapper2(mapper1(x))
         break
     else:
@@ -125,7 +148,7 @@ def test_result_map_error_piped(msg: str, y: int):
 
     ys = xs.pipe(result.map(mapper))
 
-    for err in ys.match(Error):
+    for err in ys.match(Error[int, str]):
         assert err == msg
         break
     else:
@@ -138,7 +161,7 @@ def test_result_map_error_fluent(msg: str, y: int):
     mapper: Callable[[int], int] = lambda x: x + y
 
     ys = xs.map(mapper)
-    for err in ys.match(Error):
+    for err in ys.match(Error[int, str]):
         assert err == msg
         break
     else:
@@ -152,7 +175,7 @@ def test_result_error_chained_map(msg: str, y: int):
     mapper2: Callable[[int], int] = lambda x: x * 10
 
     ys = xs.map(mapper1).map(mapper2)
-    for err in ys.match(Error):
+    for err in ys.match(Error[int, str]):
         assert err == msg
         break
     else:
@@ -165,7 +188,7 @@ def test_result_bind_piped(x: int, y: int):
     mapper: Callable[[int], Result[int, str]] = lambda x: Ok(x + y)
 
     ys = xs.pipe(result.bind(mapper))
-    for value in ys.match(Ok):
+    for value in ys.match(Ok[int, str]):
         assert Ok(value) == mapper(x)
         break
     else:
@@ -176,7 +199,7 @@ def test_result_bind_piped(x: int, y: int):
 def test_result_traverse_ok(xs: List[int]):
     ys: List[Result[int, str]] = [Ok(x) for x in xs]
     zs = sequence(ys)
-    for value in zs.match(Ok):
+    for value in zs.match(Ok[List[int], str]):
         assert sum(value) == sum(xs)
         break
     else:
@@ -189,11 +212,11 @@ def test_result_traverse_error(xs: List[int]):
     ys: List[Result[int, str]] = [Ok(x) if i == 3 else Error(error) for x, i in enumerate(xs)]
 
     zs = sequence(ys)
-    for err in zs.match(Error):
+    for err in zs.match(Error[int, str]):
         assert err == error
 
 
-def test_result_builder_zero():
+def test_result_effect_zero():
     @effect.result
     def fn():
         while False:
@@ -203,61 +226,66 @@ def test_result_builder_zero():
         fn()
 
 
-def test_result_builder_yield_ok():
+def test_result_effect_yield_ok():
     @effect.result
     def fn() -> Generator[int, int, Optional[int]]:
         yield 42
+        return None
 
     xs = fn()
     for x in xs:
         assert x == 42
 
 
-def test_result_builder_return_ok():
+def test_result_effect_return_ok():
     @effect.result
     def fn() -> Generator[int, int, int]:
         x = yield 42
         return x
 
     xs = fn()
-    for x in xs.match(Ok):
+    for x in xs.match(Ok[int, Any]):
         assert x == 42
         break
     else:
         assert False
 
 
-def test_result_builder_yield_from_ok():
+def test_result_effect_yield_from_ok():
     @effect.result
     def fn() -> Generator[int, int, int]:
         x = yield from Ok(42)
         return x + 1
 
     xs = fn()
-    for x in xs.match(Ok):
+    for x in xs.match(Ok[int, Any]):
         assert x == 43
         break
     else:
         assert False
 
 
-def test_result_builder_yield_from_error():
+def test_result_effect_yield_from_error():
     error = "Do'h"
 
+    def mayfail() -> Result[int, str]:
+        return Error(error)
+
     @effect.result
     def fn() -> Generator[int, int, int]:
-        x = yield from Error(error)
-        return x
+        xs = mayfail()
+        x = yield from xs
+        return x + 1
 
     xs = fn()
-    for err in xs.match(Error):
+    for err in xs.match(Error[int, str]):
         assert err == error
         break
     else:
         assert False, "Should not happen"
 
 
-def test_result_builder_multiple_ok():
+def test_result_effect_multiple_ok():
     @effect.result
     def fn() -> Generator[int, int, int]:
         x = yield 42
@@ -266,8 +294,54 @@ def test_result_builder_multiple_ok():
         return x + y
 
     xs = fn()
-    for value in xs.match(Ok):
+    for value in xs.match(Ok[int, Any]):
         assert value == 85
         break
     else:
         assert False
+
+
+def test_result_effect_throws():
+    error = CustomException("this happend!")
+
+    @effect.result
+    def fn() -> Generator[int, int, int]:
+        _ = yield from Ok(42)
+        raise error
+
+    with pytest.raises(CustomException) as exc:
+        fn()
+
+    assert exc.value == error
+
+
+def test_pipeline_none():
+
+    hn = pipeline()
+
+    assert hn(42) == Ok(42)
+
+
+def test_pipeline_works():
+    fn: Callable[[int], Result[int, Exception]] = lambda x: Ok(x * 10)
+    gn: Callable[[int], Result[int, Exception]] = lambda x: Ok(x + 10)
+
+    hn = pipeline(
+        fn,
+        gn,
+    )
+
+    assert hn(42) == Ok(430)
+
+
+def test_pipeline_error():
+    error: Result[int, str] = Error("failed")
+    fn: Callable[[int], Result[int, str]] = lambda x: Ok(x * 10)
+    gn: Callable[[int], Result[int, str]] = lambda x: error
+
+    hn = pipeline(
+        fn,
+        gn,
+    )
+
+    assert hn(42) == error
